;; Generated package description from fontaine.el  -*- no-byte-compile: t -*-
(define-package "fontaine" "2.0.0" "Set font configurations using presets" '((emacs "29.1")) :commit "353e03ece05f31131bd160176b9c3c4bc2b7b365" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/fontaine")
